"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.extend = exports.extendTransform = exports.hasNativeMethod = exports.hasExpressionExtension = void 0;
const luxon_1 = require("luxon");
const ExpressionError_1 = require("../ExpressionError");
const recast_1 = require("recast");
const ArrayExtensions_1 = require("./ArrayExtensions");
const DateExtensions_1 = require("./DateExtensions");
const NumberExtensions_1 = require("./NumberExtensions");
const StringExtensions_1 = require("./StringExtensions");
const ObjectExtensions_1 = require("./ObjectExtensions");
const EXPRESSION_EXTENDER = 'extend';
function isBlank(value) {
    return value === null || value === undefined || !value;
}
function isPresent(value) {
    return !isBlank(value);
}
const EXTENSION_OBJECTS = [
    ArrayExtensions_1.arrayExtensions,
    DateExtensions_1.dateExtensions,
    NumberExtensions_1.numberExtensions,
    ObjectExtensions_1.objectExtensions,
    StringExtensions_1.stringExtensions,
];
const genericExtensions = {
    isBlank,
    isPresent,
};
const EXPRESSION_EXTENSION_METHODS = Array.from(new Set([
    ...Object.keys(StringExtensions_1.stringExtensions.functions),
    ...Object.keys(NumberExtensions_1.numberExtensions.functions),
    ...Object.keys(DateExtensions_1.dateExtensions.functions),
    ...Object.keys(ArrayExtensions_1.arrayExtensions.functions),
    ...Object.keys(ObjectExtensions_1.objectExtensions.functions),
    ...Object.keys(genericExtensions),
]));
const EXPRESSION_EXTENSION_REGEX = new RegExp(`(\\$if|\\.(${EXPRESSION_EXTENSION_METHODS.join('|')}))\\s*\\(`);
const isExpressionExtension = (str) => EXPRESSION_EXTENSION_METHODS.some((m) => m === str);
const hasExpressionExtension = (str) => EXPRESSION_EXTENSION_REGEX.test(str);
exports.hasExpressionExtension = hasExpressionExtension;
const hasNativeMethod = (method) => {
    if ((0, exports.hasExpressionExtension)(method)) {
        return false;
    }
    const methods = method
        .replace(/[^\w\s]/gi, ' ')
        .split(' ')
        .filter(Boolean);
    return methods.every((methodName) => {
        return [String.prototype, Array.prototype, Number.prototype, Date.prototype].some((nativeType) => {
            if (methodName in nativeType) {
                return true;
            }
            return false;
        });
    });
};
exports.hasNativeMethod = hasNativeMethod;
const findParent = (path, matcher) => {
    let parent = path.parentPath;
    while (parent) {
        if (matcher(parent)) {
            return parent;
        }
        parent = parent.parentPath;
    }
    return;
};
const extendTransform = (expression) => {
    try {
        const ast = (0, recast_1.parse)(expression);
        (0, recast_1.visit)(ast, {
            visitIdentifier(path) {
                var _a;
                this.traverse(path);
                if (path.node.name === '$if') {
                    const callPath = findParent(path, (p) => { var _a; return ((_a = p.value) === null || _a === void 0 ? void 0 : _a.type) === 'CallExpression'; });
                    if (!callPath || ((_a = callPath.value) === null || _a === void 0 ? void 0 : _a.type) !== 'CallExpression') {
                        return;
                    }
                    if (callPath.node.arguments.length < 2) {
                        throw new ExpressionError_1.ExpressionExtensionError('$if requires at least 2 parameters: test, value_if_true[, and value_if_false]');
                    }
                    const test = callPath.node.arguments[0];
                    const consequent = callPath.node.arguments[1];
                    const alternative = callPath.node.arguments[2] === undefined
                        ? recast_1.types.builders.booleanLiteral(false)
                        : callPath.node.arguments[2];
                    callPath.replace(recast_1.types.builders.conditionalExpression(test, consequent, alternative));
                }
            },
        });
        (0, recast_1.visit)(ast, {
            visitIdentifier(path) {
                var _a, _b, _c;
                this.traverse(path);
                if (isExpressionExtension(path.node.name) &&
                    ((_b = (_a = path.parentPath) === null || _a === void 0 ? void 0 : _a.value) === null || _b === void 0 ? void 0 : _b.type) === 'MemberExpression') {
                    const callPath = findParent(path, (p) => { var _a; return ((_a = p.value) === null || _a === void 0 ? void 0 : _a.type) === 'CallExpression'; });
                    if (!callPath || ((_c = callPath.value) === null || _c === void 0 ? void 0 : _c.type) !== 'CallExpression') {
                        return;
                    }
                    callPath.replace(recast_1.types.builders.callExpression(recast_1.types.builders.identifier(EXPRESSION_EXTENDER), [
                        path.parentPath.value.object,
                        recast_1.types.builders.stringLiteral(path.node.name),
                        recast_1.types.builders.arrayExpression(callPath.node.arguments),
                    ]));
                }
            },
        });
        return (0, recast_1.print)(ast);
    }
    catch (e) {
        return;
    }
};
exports.extendTransform = extendTransform;
function isDate(input) {
    if (typeof input !== 'string' || !input.length) {
        return false;
    }
    if (!/\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{3}Z/.test(input)) {
        return false;
    }
    const d = new Date(input);
    return d instanceof Date && !isNaN(d.valueOf()) && d.toISOString() === input;
}
function extend(input, functionName, args) {
    let foundFunction;
    if (Array.isArray(input)) {
        foundFunction = ArrayExtensions_1.arrayExtensions.functions[functionName];
    }
    else if (isDate(input) && functionName !== 'toDate') {
        input = new Date(input);
        foundFunction = DateExtensions_1.dateExtensions.functions[functionName];
    }
    else if (typeof input === 'string') {
        foundFunction = StringExtensions_1.stringExtensions.functions[functionName];
    }
    else if (typeof input === 'number') {
        foundFunction = NumberExtensions_1.numberExtensions.functions[functionName];
    }
    else if (input && (luxon_1.DateTime.isDateTime(input) || input instanceof Date)) {
        foundFunction = DateExtensions_1.dateExtensions.functions[functionName];
    }
    else if (input !== null && typeof input === 'object') {
        foundFunction = ObjectExtensions_1.objectExtensions.functions[functionName];
    }
    if (!foundFunction) {
        const inputAny = input;
        if (inputAny &&
            functionName &&
            typeof inputAny[functionName] === 'function') {
            return inputAny[functionName](...args);
        }
        foundFunction = genericExtensions[functionName];
    }
    if (!foundFunction) {
        const haveFunction = EXTENSION_OBJECTS.filter((v) => functionName in v.functions);
        if (!haveFunction.length) {
            throw new ExpressionError_1.ExpressionExtensionError(`Unknown expression function: ${functionName}`);
        }
        if (haveFunction.length > 1) {
            const lastType = `"${haveFunction.pop().typeName}"`;
            const typeNames = `${haveFunction.map((v) => `"${v.typeName}"`).join(', ')}, and ${lastType}`;
            throw new ExpressionError_1.ExpressionExtensionError(`${functionName}() is only callable on types ${typeNames}`);
        }
        else {
            throw new ExpressionError_1.ExpressionExtensionError(`${functionName}() is only callable on type "${haveFunction[0].typeName}"`);
        }
    }
    return foundFunction(input, args);
}
exports.extend = extend;
//# sourceMappingURL=ExpressionExtension.js.map